class TODO(NotImplementedError):
    def __init__(self, msg: str = "todo"):
        super().__init__(msg)


class Unreachable(NotImplementedError):
    def __init__(self, msg: str = "unimplemented"):
        super().__init__(msg)



import typing as t


class MDPosition(t.TypedDict):
    line: int
    column: int
    offset: int


class MDPositionRange(t.TypedDict):
    start: MDPosition
    end: MDPosition


class MDChildNode(t.TypedDict, total=False):
    type: str  # General identifier for node types
    children: list["MDChildNode"]  # Recursive children of any node type
    position: MDPositionRange
    value: str  # Optional, for nodes like text that hold a value
    depth: int  # Optional, for nodes like headings that have a depth
    url: t.NotRequired[str]


class MDRootNode(t.TypedDict):
    type: t.Literal["root"]  # Constrains to 'root' for the root node
    children: list[MDChildNode]  # Allows any ChildNode type in children
    position: MDPositionRange


class EditorChildData(t.TypedDict, total=False):
    text: str
    items: list["EditorChildNode"]


class EditorChildNode(t.TypedDict):
    type: str
    data: EditorChildData


class EditorRootNode(t.TypedDict):
    time: int
    blocks: list[EditorChildNode]
    version: str



"""
mdast to editorjs
"""

import abc
import re
import typing as t
from html.parser import HTMLParser
from urllib.parse import urlparse

import humanize
import markdown2

from .exceptions import TODO, Unreachable
from .types import EditorChildData, MDChildNode
import html2markdown


class EditorJSBlock(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def to_markdown(cls, data: EditorChildData) -> str: ...

    @classmethod
    @abc.abstractmethod
    def to_json(cls, node: MDChildNode) -> list[dict]: ...

    @classmethod
    @abc.abstractmethod
    def to_text(cls, node: MDChildNode) -> str: ...


BLOCKS: dict[str, EditorJSBlock] = {}


def block(*names: str):
    def wrapper(cls):
        for name in names:
            BLOCKS[name] = cls
        return cls

    return wrapper


def process_styled_content(item: MDChildNode, strict: bool = True) -> str:
    """
    Processes styled content (e.g., bold, italic) within a list item.

    Args:
        item: A ChildNode dictionary representing an inline element or text.
        strict: Raise if 'type' is not one defined in 'html_wrappers'

    Returns:
        A formatted HTML string based on the item type.
    """
    _type = item.get("type")

    print('process_styled_content', _type)

    html_wrappers = {
        "text": "{value}",
        "html": "{value}",
        "emphasis": "<i>{value}</i>",
        "strong": "<b>{value}</b>",
        "strongEmphasis": "<b><i>{value}</i></b>",
        "link": '<a href="{url}">{value}</a>',
        "inlineCode": '<code class="inline-code">{value}</code>',
    }

    if _type in BLOCKS:
        return BLOCKS[_type].to_text(item)

    if strict and _type not in html_wrappers:
        raise ValueError(f"Unsupported type {_type} in paragraph")

    # Process children recursively if they exist, otherwise use the direct value
    if children := item.get("children"):
        value = "".join(process_styled_content(child) for child in children)
    else:
        value = item.get("value", "")

    template = html_wrappers.get(_type, "{value}")
    return template.format(
        value=value, url=item.get("url", ""), caption=item.get("caption", "")
    )


def default_to_text(node: MDChildNode):
    if node["type"] == "paragraph":
        return "".join(
            process_styled_content(child) for child in node.get("children", [])
        )
    else:
        return process_styled_content(node)

    # return "".join(
    #     process_styled_content(child) for child in node.get("children", [])
    # ) or process_styled_content(node)


@block("heading", "header")
class HeadingBlock(EditorJSBlock):
    @classmethod
    def to_markdown(cls, data: EditorChildData) -> str:
        level = data.get("level", 1)
        text = data.get("text", "")
        tunes = data.get("tunes", {})

        if not (1 <= level <= 6):
            raise ValueError("Header level must be between 1 and 6.")

        if (
                tunes.get("alignmentTune")
                and (alignment := tunes["alignmentTune"].get("alignment"))
                and (alignment != "left")
        ):
            # can't just return regular HTML because then it will turn into a raw block
            return AlignmentBlock.to_markdown(
                {
                    "text": text,
                    "tag": f"h{level}",
                    "alignment": alignment,
                }
            )

        return f"{'#' * level} {text}\n"

    @classmethod
    def to_json(cls, node: MDChildNode) -> list[dict]:
        """
        Converts a Markdown header block into structured block data.

        Args:
            node: A RootNode dictionary with 'depth' and 'children'.

        Returns:
            A ChildNode dictionary representing the header data, or None if no children exist.

        Raises:
            ValueError: If an unsupported heading depth is provided.
        """

        depth = node.get("depth")

        if depth is None or not (1 <= depth <= 6):
            raise ValueError("Heading depth must be between 1 and 6.")

        return [{"data": {"level": depth, "text": cls.to_text(node)}, "type": "header"}]

    @classmethod
    def to_text(cls, node: MDChildNode) -> str:
        children = node.get("children", [])
        if children is None or not len(children) == 1:
            raise ValueError("Header block must have exactly one child element")
        child = children[0]
        return child.get("value", "")


def paragraph_block(text: str):
    return {"type": "paragraph", "data": {"text": text}}


def raw_block(html: str):
    return {"type": "raw", "data": {"html": html}}


@block("paragraph")
class ParagraphBlock(EditorJSBlock):
    @classmethod
    def to_markdown(cls, data: EditorChildData) -> str:
        text = data.get("text", "")
        tunes = data.get("tunes", {})

        if (
                tunes.get("alignmentTune")
                and (alignment := tunes["alignmentTune"].get("alignment"))
                and (alignment != "left")
        ):
            return AlignmentBlock.to_markdown(
                {
                    "text": text,
                    "tag": "p",
                    "alignment": alignment,
                }
            )

        # deal with bold etc:
        text = html2markdown.convert(text)

        return f"{text}\n\n"

    @classmethod
    def to_json(cls, node: MDChildNode) -> list[dict]:
        result = []
        current_text = ""

        skip = 0
        nodes = node.get("children", [])
        any_html = False

        for idx, child in enumerate(nodes):
            if skip:
                skip -= 1
                continue

            _type = child.get("type")
            any_html |= _type == "html"

            # deal with custom types
            if _type == "html" and child.get("value", "").startswith("<editorjs"):
                # special type, e.g. <editorjs type="linkTool" href=...>...</editorjs>

                if child.get("value", "").endswith("/>"):
                    # self-closing
                    result.append(EditorJSCustom.to_json(node))
                else:
                    # <editorjs>something</editorjs> = 3 children
                    result.extend(
                        EditorJSCustom.to_json({"children": nodes[idx: idx + 2]})
                    )

                    skip = 2

                continue

            elif _type == "image":
                if current_text:
                    # {"id":"zksvpxQTDD","type":"raw","data":{"html":"<marquee> raw </marquee>"}}
                    result.append(
                        raw_block(current_text)
                        if any_html
                        else paragraph_block(current_text)
                    )
                    current_text = ""
                    any_html = False  # reset

                result.extend(ImageBlock.to_json(child))
            else:
                child_text = cls.to_text(child)
                _child_text = child_text.strip()
                if _child_text.startswith("|") and _child_text.endswith("|"):
                    # note: this just supports text-only tables.
                    # tables with more complex elements break into multiple children.
                    # and mdast DOES support converting into a Table/TableCell structure
                    # via the GFM exttension
                    # but their default mdast->md converter does NOT implement these functionalities.
                    result.extend(TableBlock.to_json(child))
                    continue

                current_text += child_text

        # final text after image:
        if current_text:
            result.append(
                raw_block(current_text) if any_html else paragraph_block(current_text)
            )

        return result

    @classmethod
    def to_text(cls, node: MDChildNode) -> str:
        return default_to_text(node)


@block("list")
class ListBlock(EditorJSBlock):
    @classmethod
    def to_markdown(cls, data: EditorChildData) -> str:
        style = data.get("style", "unordered")
        items = data.get("items", [])

        def parse_items(subitems: list[dict[str, t.Any]], depth: int = 0) -> str:
            markdown_items = []
            for index, item in enumerate(subitems):
                prefix = f"{index + 1}." if style == "ordered" else "-"
                line = f"{'\t' * depth}{prefix} {item['content']}"
                markdown_items.append(line)

                # Recurse if there are nested items
                if item.get("items"):
                    markdown_items.append(parse_items(item["items"], depth + 1))

            return "\n".join(markdown_items)

        return "\n" + parse_items(items) + "\n\n"

    @classmethod
    def to_json(cls, node: MDChildNode) -> list[dict]:
        """
        Converts a Markdown list block with nested items and styling into structured block data.

        Args:
            node: A RootNode dictionary with 'ordered' and 'children'.

        Returns:
            A dictionary representing the structured list data with 'items' and 'style'.
        """
        items = []
        # checklists are not supported (well) by mdast
        # so we detect it ourselves:
        could_be_checklist = True

        def is_checklist(value: str) -> bool:
            return value.strip().startswith(("[ ]", "[x]"))

        for child in node["children"]:
            content = ""
            subitems = []
            # child can have content and/or items
            for grandchild in child["children"]:
                _type = grandchild.get("type", "")
                if _type == "paragraph":
                    subcontent = ParagraphBlock.to_text(grandchild)
                    could_be_checklist = could_be_checklist and is_checklist(subcontent)
                    content += "" + subcontent
                elif _type == "list":
                    could_be_checklist = False
                    subitems.extend(ListBlock.to_json(grandchild)[0]["data"]["items"])
                else:
                    raise ValueError(f"Unsupported type {_type} in list")

            items.append(
                {
                    "content": content,
                    "items": subitems,
                }
            )

        if could_be_checklist:
            return [
                {
                    "type": "checklist",
                    "data": {
                        "items": [
                            {
                                "text": x["content"]
                                .removeprefix("[ ] ")
                                .removeprefix("[x] "),
                                "checked": x["content"].startswith("[x]"),
                            }
                            for x in items
                        ]
                    },
                }
            ]
        else:
            return [
                {
                    "data": {
                        "items": items,
                        "style": "ordered" if node.get("ordered") else "unordered",
                    },
                    "type": "list",
                }
            ]

    @classmethod
    def to_text(cls, node: MDChildNode) -> str:
        return ""


@block("checklist")
class ChecklistBlock(ListBlock):
    @classmethod
    def to_markdown(cls, data: EditorChildData) -> str:
        markdown_items = []

        for item in data.get("items", []):
            text = item.get("text", "").strip()
            char = "x" if item.get("checked", False) else " "
            markdown_items.append(f"- [{char}] {text}")

        return "\n" + "\n".join(markdown_items) + "\n\n"


@block("thematicBreak", "delimiter")
class DelimiterBlock(EditorJSBlock):
    @classmethod
    def to_markdown(cls, data: EditorChildData) -> str:
        return "***\n"

    @classmethod
    def to_json(cls, node: MDChildNode) -> list[dict]:
        return [
            {
                "type": "delimiter",
                "data": {},
            }
        ]

    @classmethod
    def to_text(cls, node: MDChildNode) -> str:
        return ""


@block("code")
class CodeBlock(EditorJSBlock):
    @classmethod
    def to_markdown(cls, data: EditorChildData) -> str:
        code = data.get("code", "")
        return f"```\n" f"{code}" f"\n```\n"

    @classmethod
    def to_json(cls, node: MDChildNode) -> list[dict]:
        return [
            {
                "data": {"code": cls.to_text(node)},
                "type": "code",
            }
        ]

    @classmethod
    def to_text(cls, node: MDChildNode) -> str:
        return node.get("value", "")


@block("image")
class ImageBlock(EditorJSBlock):
    @classmethod
    def to_markdown(cls, data: EditorChildData) -> str:
        url = data.get("url", "") or data.get("file", {}).get("url", "")
        caption = data.get("caption", "")

        with_border = "1" if data.get("withBorder") else ""
        with_background = "1" if data.get("withBackground") else ""
        stretched = "1" if data.get("stretched") else ""

        # always custom type so we can render as <figure> instead of markdown2's default (simple <img>)
        return f"""<editorjs type="image" caption="{caption}" border="{with_border}" background="{with_background}" stretched="{stretched}" url="{url}" />\n\n"""

    @classmethod
    def _caption(cls, node: MDChildNode):
        return node.get("alt") or node.get("caption") or ""

    @classmethod
    def to_json(cls, node: MDChildNode) -> list[dict]:
        return [
            {
                "type": "image",
                "data": {
                    "file": {"url": node.get("url")},
                    "caption": cls._caption(node),
                    "withBorder": bool(node.get("border", False)),
                    "stretched": bool(node.get("stretched", False)),
                    "withBackground": bool(node.get("background", False)),
                },
            }
        ]

    @classmethod
    def to_text(cls, node: MDChildNode) -> str:
        caption = cls._caption(node)
        url = node.get("url")

        background = node.get("background") or ""
        stretched = node.get("stretched") or ""
        border = node.get("border") or ""

        return f"""
        <div class="ce-block {stretched and 'ce-block--stretched'}">
            <div class="ce-block__content">
            <div class="cdx-block image-tool image-tool--filled {background and 'image-tool--withBackground'} {stretched and 'image-tool--stretched'} {border and 'image-tool--withBorder'}">
                <div class="image-tool__image">
                    <figure>
                        <img class="image-tool__image-picture" src="{url}" title="{caption}" alt="{caption}">
                        <figcaption>{caption}</figcaption>
                    </figure>
                </div>
            </div>
        </div>
        """


@block("blockquote", "quote")
class QuoteBlock(EditorJSBlock):
    re_cite = re.compile(r"<cite>(.+?)<\/cite>")

    @classmethod
    def to_markdown(cls, data: EditorChildData) -> str:
        text = data.get("text", "")
        result = f"> {text}\n"
        if caption := data.get("caption", ""):
            result += f"> <cite>{caption}</cite>\n"
        return result + "\n"

    @classmethod
    def to_json(cls, node: MDChildNode) -> list[dict]:
        caption = ""
        text = cls.to_text(node).replace("\n", "<br/>\n")

        if cite := re.search(cls.re_cite, text):
            # Capture the value of the first group
            caption = cite.group(1)
            # Remove the <cite>...</cite> tags from the text
            text = re.sub(cls.re_cite, "", text)

        return [
            {
                "data": {
                    "alignment": "left",
                    "caption": caption,
                    "text": text,
                },
                "type": "quote",
            }
        ]

    @classmethod
    def to_text(cls, node: MDChildNode) -> str:
        return "".join(
            process_styled_content(child) for child in node.get("children", [])
        )


@block("raw", "html")
class RawBlock(EditorJSBlock):

    @classmethod
    def to_markdown(cls, data: EditorChildData) -> str:
        text = data.get("html", "")
        return f"{text}\n\n"

    @classmethod
    def to_json(cls, node: MDChildNode) -> list[dict]:
        raw = cls.to_text(node)

        if raw.startswith("<editorjs"):
            # not a raw block but (probably) a self-closing editorjs block
            return EditorJSCustom.to_json({"children": [node]})
        else:
            return [raw_block(raw)]

    @classmethod
    def to_text(cls, node: MDChildNode) -> str:
        return node.get("value", "")


@block("table")
class TableBlock(EditorJSBlock):

    @classmethod
    def to_markdown(cls, data: EditorChildData) -> str:
        """
        | Script | Interpreter | User | System |   |
        |--------|-------------|------|--------|---|
        |        |             |      |        |   |
        |        |             |      |        |   |
        |        |             |      |        |   |
        """

        table = ""
        rows = data.get("content", [])

        # Add an empty header row if no headings are provided
        if not data.get("withHeadings", False) and rows:
            table += "| " + " | ".join([""] * len(rows[0])) + " |\n"
            table += "|" + " - |" * len(rows[0]) + "\n"

        # Populate rows
        for idx, tr in enumerate(rows):
            table += "| " + " | ".join(tr) + " |\n"

            # Add separator if headings are enabled and it's the first row
            if not idx and data.get("withHeadings", False):
                table += "|" + " - |" * len(tr) + "\n"

        return f"\n{table}\n"

    @classmethod
    def to_json(cls, node: MDChildNode) -> list[dict]:
        # content":[["Yeah","Okay"],["<i>1</i>","<code class=\"inline-code\">2</code>"]]}}]
        table = []
        with_headings = False

        # first row is headings or empty. If not empty, withHeadings is True
        # second row must be ignored
        for idx, row in enumerate(node.get("value", "").strip().split("\n")):
            tr = [_.strip() for _ in row.split("|")[1:-1]]
            if not idx:
                # first
                if any(tr):
                    with_headings = True
                    table.append(tr)

            elif idx == 1:
                continue
            else:
                table.append(tr)

        return [
            {
                "type": "table",
                "data": {
                    "content": table,
                    "withHeadings": with_headings,
                },
            }
        ]

    @classmethod
    def to_text(cls, node: MDChildNode) -> str:
        # I think this might be triggered if there is a table (deeply) within a paragraph block?
        raise TODO(["TableBlock.to_text", node])


@block("linkTool")
class LinkBlock(EditorJSBlock):
    @classmethod
    def to_markdown(cls, data: EditorChildData) -> str:
        link = data.get("link", "")
        meta = data.get("meta", {})
        title = meta.get("title", "")
        description = meta.get("description", "")
        image = meta.get("image", {}).get("url", "")
        return f"""<editorjs type="linkTool" href="{link}" title="{title}" image="{image}">{description}</editorjs>\n\n"""

    @classmethod
    def to_json(cls, node: MDChildNode) -> list[dict]:
        return [
            {
                "type": "linkTool",
                "data": {
                    "link": node.get("href", ""),
                    "meta": {
                        "title": node.get("title", ""),
                        "description": node.get("body", ""),
                        "image": {
                            "url": node.get("image", ""),
                        },
                    },
                },
            }
        ]

    @classmethod
    def to_text(cls, node: MDChildNode) -> str:
        url = node.get("href", "")
        image = node.get("image", "")
        title = node.get("title", "")
        body = node.get("body", "")
        domain = urlparse(url).netloc

        return f"""
        <div class="link-tool">
            <a class="link-tool__content link-tool__content--rendered" target="_blank"
               rel="nofollow noindex noreferrer" href="{url}">
                <div class="link-tool__image"
                     style="background-image: url(&quot;{image}&quot;);"></div>
                <div class="link-tool__title">{title}</div>
                <p class="link-tool__description">{body}</p>
                <span class="link-tool__anchor">{domain}</span>
            </a>
        </div>
        """


@block("attaches")
class AttachmentBlock(EditorJSBlock):

    @classmethod
    def to_markdown(cls, data: EditorChildData) -> str:
        title = data.get("title", "")
        file = data.get("file", {})
        url = file.get("url", "")
        name = file.get("name", "")
        extension = file.get("extension", "")
        size = file.get("size", "")

        return f"""<editorjs type="attaches" file="{url}" name="{name}" extension="{extension}" size="{size}">{title}</editorjs>\n\n"""

    @classmethod
    def to_json(cls, node: MDChildNode) -> list[dict]:
        return [
            {
                "type": "attaches",
                "data": {
                    "file": {
                        "url": node.get("file", ""),
                        "name": node.get("name", ""),
                        "extension": node.get("extension", ""),
                        "size": node.get("size", ""),
                    },
                    "title": node.get("body", ""),
                },
            }
        ]

    @classmethod
    def to_text(cls, node: MDChildNode) -> str:
        # {'type': 'attaches', 'file': 'https://py4web.leiden.dockers.local/img/upload/8.deb?hash=778760cf05483147b2ff0fa0ddeab2b22d9343e8', 'name': 'gemistdownloadermd5-08d0e3cdb4f7e81986bdd0c60294dec03.0.0.5-1.deb', 'extension': 'deb', 'size': '1613660', 'body': 'gemistdownloader...

        extension = node.get("extension", "")

        file_icon = (
            f"""
            <div class="cdx-attaches__file-icon-background"></div>
            <div class="cdx-attaches__file-icon-label" title="{extension}">{extension}</div>
            """
            if extension
            else """
            <div class="cdx-attaches__file-icon-background">
                <svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" fill="none" viewBox="0 0 24 24"><path stroke="currentColor" stroke-linecap="round" stroke-linejoin="round" stroke-width="2" d="M13.3236 8.43554L9.49533 12.1908C9.13119 12.5505 8.93118 13.043 8.9393 13.5598C8.94741 14.0767 9.163 14.5757 9.53862 14.947C9.91424 15.3182 10.4191 15.5314 10.9422 15.5397C11.4653 15.5479 11.9637 15.3504 12.3279 14.9908L16.1562 11.2355C16.8845 10.5161 17.2845 9.53123 17.2682 8.4975C17.252 7.46376 16.8208 6.46583 16.0696 5.72324C15.3184 4.98066 14.3086 4.55425 13.2624 4.53782C12.2162 4.52138 11.2193 4.91627 10.4911 5.63562L6.66277 9.39093C5.57035 10.4699 4.97032 11.9473 4.99467 13.4979C5.01903 15.0485 5.66578 16.5454 6.79264 17.6592C7.9195 18.7731 9.43417 19.4127 11.0034 19.4374C12.5727 19.462 14.068 18.8697 15.1604 17.7907L18.9887 14.0354"></path></svg>
            </div>
            """
        )

        file_size = (
            f"""<div class="cdx-attaches__size">{humanize.naturalsize(int(size))}</div>"""
            if (size := node.get("size", ""))
            else ""
        )

        return f"""
        <div class="cdx-attaches cdx-attaches--with-file">
            <div class="cdx-attaches__file-icon">
                {file_icon}
            </div>
            <div class="cdx-attaches__file-info">
                <div class="cdx-attaches__title" data-placeholder="File title" data-empty="false">
                    {node.get("body", "")}
                </div>
                {file_size}
            </div>
            <a class="cdx-attaches__download-button" href="{node.get('file', '')}" target="_blank" rel="nofollow noindex noreferrer" title="{node.get('name', '')}">
                <svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" fill="none" viewBox="0 0 24 24"><path stroke="currentColor" stroke-linecap="round" stroke-width="2" d="M7 10L11.8586 14.8586C11.9367 14.9367 12.0633 14.9367 12.1414 14.8586L17 10"></path></svg>
            </a>
        </div>
        """


@block("alignment")
class AlignmentBlock(EditorJSBlock):
    @classmethod
    def to_markdown(cls, data: EditorChildData) -> str:
        tag = data["tag"]
        alignment = data["alignment"]
        content = data["text"]

        return f"<editorjs type='alignment' tag='{tag}' alignment='{alignment}'>{content}</editorjs>\n\n"

    @classmethod
    def to_json(cls, node: MDChildNode) -> list[dict]:
        # only paragraph and headers can be aligned
        tag: str = node["tag"]
        text: str = node["body"]
        alignment = node["alignment"]
        data = {"text": text}

        if tag == "p":
            _type = "paragraph"
        elif tag.startswith("h"):
            _type = "header"
            data["level"] = int(tag.removeprefix("h"))
        else:
            # doesn't support alignment
            raise NotImplementedError(f"Unsupported tag for alignment: {tag}")

        return [
            {
                "type": _type,
                "data": data,
                "tunes": {"alignmentTune": {"alignment": alignment}},
            }
        ]

    @classmethod
    def to_text(cls, node: MDChildNode) -> str:
        tag = node["tag"]
        body = node["body"]
        alignment = node["alignment"]
        return f"<{tag} style='text-align: {alignment}'>{body}</{tag}>"


@block("embed")
class EmbedBlock(EditorJSBlock):

    @classmethod
    def to_markdown(cls, data: EditorChildData) -> str:
        service = data.get("service", "")
        source = data.get("source", "")
        embed = data.get("embed", "")
        caption = data.get("caption", "")

        return f"<editorjs type='embed' service='{service}' source='{source}' embed='{embed}' caption='{caption}'/>\n\n"

    @classmethod
    def to_json(cls, node: MDChildNode) -> list[dict]:
        return [{"type": "embed", "data": node}]

    @classmethod
    def to_text(cls, node: MDChildNode) -> str:
        source = node.get("source", "")
        embed = node.get("embed", "")
        caption = node.get("caption", "")
        return f"""
        <div class="cdx-block embed-tool">
            <iframe title='{caption}' style="width:100%;" height="320" frameborder="0" allowfullscreen="" src="{embed}" class="embed-tool__content"></iframe>
        </div>
        """


### end blocks


class AttributeParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.attributes = {}
        self.data = None

    def handle_starttag(self, tag, attrs):
        # Collect attributes when the tag is encountered
        self.attributes = dict(attrs)

    def handle_data(self, data):
        self.data = data


class EditorJSCustom(EditorJSBlock, markdown2.Extra):
    """
    Special type of block to deal with custom attributes.

    This is both a special editorjs block as well as a markdown2 plugin!
    """

    name = "editorjs"
    order = (), (markdown2.Stage.POSTPROCESS,)

    @classmethod
    def parse_html(cls, html: str):
        parser = AttributeParser()
        parser.feed(html)

        return parser.attributes, parser.data

    @classmethod
    def to_markdown(cls, data: EditorChildData) -> str:
        raise Unreachable("Custom Blocks have their own to_markdown logic.")

    @classmethod
    def _find_right_block(cls, html: str) -> tuple[EditorJSBlock, dict]:
        attrs, body = cls.parse_html(html)
        _type = attrs.get("type", "")
        attrs.setdefault("body", body)  # only if there is no such attribute yet

        handler = BLOCKS.get(_type)

        if not handler:
            raise ValueError(f"Unknown custom type {_type}")

        return handler, attrs

    @classmethod
    def to_json(cls, node: MDChildNode) -> list[dict]:
        html = "".join(_["value"] for _ in node.get("children", []))
        handler, attrs = cls._find_right_block(html)
        return handler.to_json(attrs)

    @classmethod
    def to_text(cls, node: MDChildNode) -> str:
        handler, attrs = cls._find_right_block(node.get("value", ""))
        return handler.to_text(attrs)

    # markdown2:
    re_short = re.compile(r"<editorjs.*?/>")
    re_long = re.compile(r"<editorjs.*?>.*?</editorjs>")

    def run(self, text: str) -> str:
        def replace_html(match):
            handler, attrs = self._find_right_block(match.group())
            return handler.to_text(attrs)

        # Substitute using the replacement functions
        text = self.re_long.sub(replace_html, text)
        text = self.re_short.sub(replace_html, text)

        return text


EditorJSCustom.register()



import json
import textwrap

from editorjs import EditorJS

EXAMPLE_MD = textwrap.dedent("""
    # Heading

    Paragraph with *special* content __yeah__ such as *[links](https://g.co)!*

    - list (unordered)
    - with
        - *nested* lvl 1.0
        - __nested__ lvl 1.1
            - nested lvl 2.0
    - [items](https://g.co)

    1. another
    2. list (ordered)
        1. with 
        2. nesting
    3. *and styling*

    ***

    Checkbox:
    - [ ] unchecked
    - [ ] *unchecked*
    - [x] checked
    - [x] *checked*

    Inline `code`

    ```python
    def real_code(example: str): 
        pass
    ```

    Image: ![Caption](https://py4web.leiden.dockers.local/img/upload/16.txt?hash=3f4b66f9be3306fa7f42c0bcf6238d1101d9c75f "Caption"); post image


    > The secret to creativity is knowing how to hide your sources.
    > Regel 2
    > <cite>Albert Einstein</cite>
    """)

EXAMPLE_JSON = r"""{"time":1730905327104,"blocks":[{"id":"1pY2hYfB1A","type":"header","data":{"text":"Heading","level":1}},{"id":"akvQa8k7ne","type":"paragraph","data":{"text":"Paragraph with special content yeah such as <a href=\"https://g.co\">links</a>!"}},{"id":"mtXTKYuP8k","type":"list","data":{"style":"unordered","items":[{"content":"list (unordered)","items":[]},{"content":"with","items":[{"content":"<i>nested</i> lvl 1.0","items":[]},{"content":"<b>nested</b> lvl 1.1","items":[{"content":"nested lvl 2.0","items":[]}]}]},{"content":"<a href=\"https://g.co\">items</a>","items":[]}]}},{"id":"CwdqcNPGp2","type":"list","data":{"style":"ordered","items":[{"content":"another","items":[]},{"content":"list (ordered)\n2.1. with\n2.2. nesting","items":[]},{"content":"<i>and styling</i>","items":[]}]}},{"id":"4aavopqblu","type":"delimiter","data":{}},{"id":"GgTUixlCYO","type":"paragraph","data":{"text":"Checkbox:"}},{"id":"E12TkYu5E_","type":"checklist","data":{"items":[{"text":"unchecked","checked":false},{"text":"<i>unchecked</i>","checked":false},{"text":"checked","checked":true},{"text":"<i>checked</i>","checked":true}]}},{"id":"I-HNbkhXE-","type":"paragraph","data":{"text":"Inline <code class=\"inline-code\">code</code>"}},{"id":"fNKmTrZvXn","type":"code","data":{"code":"def real_code(example: str): \n    pass"}},{"id":"bfHSwu6hjy","type":"paragraph","data":{"text":"Image: "}},{"id":"gClWIHSnxn","type":"image","data":{"caption":"Caption","withBorder":false,"withBackground":false,"stretched":false,"file":{"url":"https://py4web.leiden.dockers.local/img/upload/16.txt?hash=3f4b66f9be3306fa7f42c0bcf6238d1101d9c75f"}}},{"id":"NZ_ruz2mXL","type":"paragraph","data":{"text":"; post image"}},{"id":"Nn7p_3Whpy","type":"quote","data":{"text":"The secret to creativity is knowing how to hide your sources.<br>\nRegel 2<br>\n","caption":"Albert Einstein","alignment":"left"}}],"version":"2.30.6"}"""


def test_md():
    e = EditorJS.from_markdown(EXAMPLE_MD)

    print(e.to_mdast())
    print(e.to_markdown())
    print(e.to_json())
    print(e.to_html())


def test_json():
    e = EditorJS.from_json(EXAMPLE_JSON)

    print(e.to_mdast())
    print(e.to_json())
    print(e.to_markdown())
    print(e.to_html())


# def test_lossless():
#     e = EditorJS.from_markdown(EXAMPLE_MD)
#     assert e == EditorJS.from_mdast(e.to_mdast())
#     assert e == EditorJS.from_json(e.to_json())
#     assert e == EditorJS.from_markdown(e.to_markdown())
#
#     e = EditorJS.from_json(EXAMPLE_JSON)
#     assert e == EditorJS.from_mdast(e.to_mdast())
#     assert e == EditorJS.from_json(e.to_json())
#     assert e == EditorJS.from_markdown(e.to_markdown())

LINKTOOL_JSON = r"""{"time":1730911265307,"blocks":[{"id":"A07WMZn2iv","type":"linkTool","data":{"link":"https://fb.me","meta":{"title":"","description":"Meld je aan bij Facebook om te delen en contact te maken met je vrienden, familie en mensen die je kent.","image":{"url":"https://www.facebook.com/images/fb_icon_325x325.png"}}}}],"version":"2.30.6"}"""


def test_linktool():
    e = EditorJS.from_json(LINKTOOL_JSON)
    print(e.to_html())
    print(e.to_json())
    print(e.to_markdown())
    print(e.to_html())


TABLE1_JSON = r"""{"time":1730984047714,"blocks":[{"id":"SNXL5vru_a","type":"table","data":{"withHeadings":false,"stretched":false,"content":[["1.1","2.1"],["1.2","2.2"]]}},{"id":"q0IC_sL8P5","type":"paragraph","data":{"text":"<mark class=\"cdx-marker\">marked</mark>"}}],"version":"2.30.6"}"""
TABLE2_JSON = r"""{"time":1730984305796,"blocks":[{"id":"vBf5hT3jeR","type":"linkTool","data":{"link":"https://fb.me","meta":{"title":"","description":"Meld je aan bij Facebook om te delen en contact te maken met je vrienden, familie en mensen die je kent.","image":{"url":"https://www.facebook.com/images/fb_icon_325x325.png"}}}},{"id":"7bP-0bw1OT","type":"table","data":{"withHeadings":true,"stretched":false,"content":[["Yeah","Okay"],["<i>1</i>","<code class=\"inline-code\">2</code>"]]}}],"version":"2.30.6"}"""


def test_table():
    e = EditorJS.from_json(TABLE1_JSON)
    print(e.to_markdown())
    print(e.to_json())
    print(e.to_html())

    e = EditorJS.from_json(TABLE2_JSON)
    print(e.to_markdown())
    print(e.to_json())
    print(e.to_html())


ATTACHMENT_JSON = r"""{"time":1730989152705,"blocks":[{"id":"s3k3WCp1da","type":"header","data":{"text":"Hi","level":2}},{"id":"EehvyuBEKx","type":"paragraph","data":{"text":"<b>Paragraph</b> <i>met</i> <a href=\"http://g.co\">mooie</a> <code class=\"inline-code\">markup </code><mark class=\"cdx-marker\">asdf</mark>"}},{"id":"S1ENjQ6XEw","type":"list","data":{"style":"unordered","items":[{"content":"u","items":[]},{"content":"l","items":[]}]}},{"id":"ezy6f8-7gj","type":"list","data":{"style":"ordered","items":[{"content":"o","items":[]},{"content":"l","items":[]}]}},{"id":"DsHJI0tccS","type":"image","data":{"caption":"streamers","withBorder":false,"withBackground":false,"stretched":false,"file":{"url":"https://py4web.leiden.dockers.local/img/upload/17.txt?hash=3ef2d77e2f4f493651119f65d1657b01de939adb"}}},{"id":"VhvLUtaKcB","type":"quote","data":{"text":"ik ben een streamer","caption":"Frank Lammers","alignment":"left"}},{"id":"yrQLgdrYWC","type":"delimiter","data":{}},{"id":"88cMVCXNZF","type":"table","data":{"withHeadings":true,"stretched":false,"content":[["Tafels","Deuren"],["1 * 1","1"],["2 * 2","4"]]}},{"id":"vlSOey3tKF","type":"code","data":{"code":"if crash: don't"}},{"id":"zPzINXsico","type":"raw","data":{"html":"<marquee>Wheeeee</marquee>"}},{"id":"rbc6buobti","type":"checklist","data":{"items":[{"text":"check","checked":false},{"text":"one","checked":true},{"text":"two","checked":false}]}},{"id":"0JGAjTs_77","type":"linkTool","data":{"link":"https://trialandsuccess.nl","meta":{"title":"Trial and Success","description":"Welkom op de website van Trial and Success! Ik, Robin van der Noord, ontwerp en bouw websites en webapplicaties, helemaal toegespitst op de wensen van de klant. Ook kan ik de hosting en het onderhoud van uw website verzorgen.","image":{"url":"https://trialandsuccess.nl/static/images/og.png"}}}},{"id":"h8Zt0aKD5p","type":"attaches","data":{"file":{"url":"https://py4web.leiden.dockers.local/img/upload/18.txt?hash=f114bad0e27c84eb1f70ba2d6168d2a76b9efca9"},"title":"Attachment, Download nu!"}}],"version":"2.30.6"}"""


def test_attachment():
    e = EditorJS.from_json(ATTACHMENT_JSON)
    print(e.to_markdown())
    print(e.to_json())
    print(e.to_html())

    attachment_two = r"""[{"id":"RJGyL0ZWOL","type":"attaches","data":{"file":{"url":"https://py4web.leiden.dockers.local/img/upload/8.deb?hash=778760cf05483147b2ff0fa0ddeab2b22d9343e8","name":"gemistdownloader_3.0.0.5-1.deb","extension":"deb","size":1613660},"title":"gemistdownloader_3.0.0.5-1"}}]"""

    e = EditorJS.from_json(attachment_two)
    print(e.to_markdown())
    print(e.to_html())
    j = json.loads(e.to_json())
    print(j)

    assert j["blocks"][0]["data"]["file"]["size"]
    assert j["blocks"][0]["data"]["file"]["extension"]


def test_raw_html():
    # e = EditorJS.from_markdown(textwrap.dedent("""
    # # Raw  HTML
    #
    # <marquee>This ain't no paragraph</marquee>
    #
    # This is a paragraph
    # """))
    #
    # blocks = json.loads(e.to_json())
    #
    # print(blocks)
    #
    # assert blocks["blocks"][1]["type"] == "raw", blocks["blocks"][1]["type"]
    # assert blocks["blocks"][2]["type"] == "paragraph", blocks["blocks"][2]["type"]
    #

    raw_html_json = r"""{"time":1730989152705,"blocks":[{"id":"DGQwRibbof","type":"paragraph","data":{"text":"The Start"}},{"id":"NrvWoJ2bVI","type":"raw","data":{"html":" <marquee>    <kaas>mannetje    </kaas>    </marquee>Einde!"}},{"id":"DGQwRibbof","type":"paragraph","data":{"text":"The End!"}}],"version":"2.30.6"}"""

    e = EditorJS.from_json(raw_html_json)
    blocks = json.loads(e.to_json())

    assert blocks["blocks"][0]["type"] == "paragraph", blocks["blocks"][0]["type"]
    assert blocks["blocks"][1]["type"] == "raw", blocks["blocks"][1]["type"]
    assert blocks["blocks"][2]["type"] == "paragraph", blocks["blocks"][2]["type"]

    raw_html_json = r"""[{"id":"xGbqrb40Uz","type":"raw","data":{"html":"<marquee><kaas>mannetje</kaas></marquee>"}}]"""

    e = EditorJS.from_json(raw_html_json)

    print(e.to_mdast(), e.to_markdown())
    e = EditorJS.from_markdown(e.to_markdown())
    print(e.to_mdast(), e.to_markdown())

    e = EditorJS.from_markdown(e.to_markdown())
    e = EditorJS.from_markdown(e.to_markdown())
    e = EditorJS.from_markdown(e.to_markdown())
    e = EditorJS.from_markdown(e.to_markdown())
    e = EditorJS.from_markdown(e.to_markdown())

    print(e.to_markdown())
    assert "     " not in e.to_markdown(), "added too many whitespaces"

    another_md = r"""<marquee> <kaas>mannetje </kaas> </marquee> <yep>pa

</yep>

sdafasdgasdgdsag

asdfsadfsdf dsfasdf

asdfsdajgdsjaklgkjds
"""

    e = EditorJS.from_markdown(another_md)
    print(e.to_markdown())
    print(e.to_json())
    assert e.to_json()


def test_code():
    e = EditorJS.from_markdown(textwrap.dedent("""
    Read code:
    
    ```
    <marquee> code </marquee>
    ```
    
    End of code
    """))

    blocks = json.loads(e.to_json())

    assert blocks["blocks"][0]["type"] == "paragraph", blocks["blocks"][0]["type"]
    assert blocks["blocks"][1]["type"] == "code", blocks["blocks"][1]["type"]
    assert blocks["blocks"][2]["type"] == "paragraph", blocks["blocks"][2]["type"]

    # check if it's still the same after export and import:

    e = EditorJS.from_json(blocks)

    blocks = json.loads(e.to_json())

    assert blocks["blocks"][0]["type"] == "paragraph", blocks["blocks"][0]["type"]
    assert blocks["blocks"][1]["type"] == "code", blocks["blocks"][1]["type"]
    assert blocks["blocks"][2]["type"] == "paragraph", blocks["blocks"][2]["type"]

    # note: without `time` and `version` boilerplate:
    blocks_json = r"""[{"id":"DD966BQf_t","type":"paragraph","data":{"text":"Pre"}},{"id":"sjr2JyuC1y","type":"code","data":{"code":"html"}},{"id":"aJN_wgBv7b","type":"paragraph","data":{"text":"Post"}}]"""

    e = EditorJS.from_json(blocks_json)

    blocks = json.loads(e.to_json())

    assert blocks["blocks"][0]["type"] == "paragraph", blocks["blocks"][0]["type"]
    assert blocks["blocks"][1]["type"] == "code", blocks["blocks"][1]["type"]
    assert blocks["blocks"][2]["type"] == "paragraph", blocks["blocks"][2]["type"]


def test_alignment():
    json_blocks = r"""[{"id":"7pEj7OVRiI","type":"header","data":{"text":"Right","level":2},"tunes":{"alignmentTune":{"alignment":"right"}}},{"id":"BxL_tpq3AD","type":"paragraph","data":{"text":"right"},"tunes":{"alignmentTune":{"alignment":"right"}}},{"id":"KOh0mMtNQf","type":"header","data":{"text":"Center","level":2},"tunes":{"alignmentTune":{"alignment":"center"}}},{"id":"AegxKSR6Oa","type":"paragraph","data":{"text":"center"},"tunes":{"alignmentTune":{"alignment":"center"}}},{"id":"xJK82ujRe5","type":"header","data":{"text":"Left","level":3},"tunes":{"alignmentTune":{"alignment":"left"}}},{"id":"bwwQwKdZf0","type":"paragraph","data":{"text":"left"},"tunes":{"alignmentTune":{"alignment":"left"}}}]"""

    e = EditorJS.from_json(json_blocks)

    print(e.to_markdown())
    print(e.to_html())

    blocks = json.loads(e.to_json())

    assert blocks["blocks"][0]["tunes"]

    print(e.to_json())


def test_embed():
    json_blocks = r"""{"time":1732710342696,"blocks":[{"id":"AzmqC3FWx3","type":"embed","data":{"service":"youtube","source":"https://www.youtube.com/watch?v=LDU_Txk06tM","embed":"https://www.youtube.com/embed/LDU_Txk06tM","width":580,"height":320,"caption":"krab"}}],"version":"2.30.6"}"""

    e = EditorJS.from_json(json_blocks)

    print(e.to_markdown())
    print(e.to_html())
    print(e.to_json())


def test_image_options():
    json_blocks = r"""{"time":1733155142016,"blocks":[{"id":"e7_WBThzLQ","type":"image","data":{"caption":"border","withBorder":true,"withBackground":false,"stretched":false,"file":{"url":"https://py4web.leiden.dockers.local/img/upload/5.jpg?hash=b39755c8a568cbf45d329e3a3128fb43065b1d1b","name":"kat.jpg","title":"kat","extension":"jpg","size":3682051}}},{"id":"B5qVcjqBuB","type":"image","data":{"caption":"stretch","withBorder":false,"withBackground":false,"stretched":true,"file":{"url":"https://py4web.leiden.dockers.local/img/upload/6.jpg?hash=b39755c8a568cbf45d329e3a3128fb43065b1d1b","name":"kat.jpg","title":"kat","extension":"jpg","size":3682051}}},{"id":"ft32yP2_cv","type":"image","data":{"caption":"background","withBorder":false,"withBackground":true,"stretched":false,"file":{"url":"https://py4web.leiden.dockers.local/img/upload/7.jpg?hash=b39755c8a568cbf45d329e3a3128fb43065b1d1b","name":"kat.jpg","title":"kat","extension":"jpg","size":3682051}}}],"version":"2.30.7"}"""

    e = EditorJS.from_json(json_blocks)

    print(e.to_markdown())
    print(e.to_html())
    print(e.to_json())


def test_figcaption():
    js = """{"time":1742471258492,"blocks":[{"id":"ZoA3rbc05C","type":"image","data":{"caption":"Party Time!","withBorder":false,"withBackground":false,"stretched":false,"file":{"url":"https://py4web.leiden.dockers.local/img/upload/23.png?hash=979795a433fc15cb94eccb3159f1f1e4054b1664"}}}],"version":"2.30.7"}"""
    e = EditorJS.from_json(js)

    html = e.to_html()

    assert "figcaption" in html

def test_bold():
    js = """{"time":1742475802066,"blocks":[{"id":"v_Kc51dnJH","type":"paragraph","data":{"text":"Deze tekst is <b>half bold</b> en half niet"},"tunes":{"alignmentTune":{"alignment":"left"}}},{"id":"q_fkuEFcY5","type":"paragraph","data":{"text":"<b>Deze tekst is heel bold</b>"},"tunes":{"alignmentTune":{"alignment":"left"}}}],"version":"2.30.7"}"""

    e = EditorJS.from_json(js)

    print(
        e._mdast,
        e.to_json(),
        e.to_html(),
        e.to_markdown(),
    )


def test_quotes():
    js = """{"time":1742477286420,"blocks":[{"id":"PpmEaxeSkq","type":"quote","data":{"text":"To baldly go where no bald man has ever gone before","caption":"a bald guy","alignment":"left"}},{"id":"26ERFQ6U3V","type":"quote","data":{"text":"Einstein was een sukkel","caption":"Einstein's ex vrouw","alignment":"left"}},{"id":"RB8AdaCd86","type":"quote","data":{"text":"Asdf","caption":"fsda-man","alignment":"left"}},{"id":"BCSus2rhUr","type":"paragraph","data":{"text":"groetjes"},"tunes":{"alignmentTune":{"alignment":"left"}}}],"version":"2.30.7"}"""

    e = EditorJS.from_json(js)

    print(
        e._mdast,
        e.to_json(),
        e.to_html(),
        e.to_markdown(),
    )




import json
import typing as t
import warnings

import markdown2
import mdast
from typing_extensions import Self

from .blocks import BLOCKS
from .exceptions import TODO
from .helpers import unix_timestamp
from .types import MDRootNode

EDITORJS_VERSION = "2.30.6"


class EditorJS:
    # internal representation is mdast, because we can convert to other types
    _mdast: MDRootNode

    def __init__(
        self,
        _mdast: str | dict,
        extras: list = ("task_list", "fenced-code-blocks", "tables", "editorjs"),
    ):
        if not isinstance(_mdast, str | dict):
            raise TypeError("Only `str` or `dict` is supported!")

        self._mdast = t.cast(
            MDRootNode, json.loads(_mdast) if isinstance(_mdast, str) else _mdast
        )

        self._md = markdown2.Markdown(extras=extras)  # todo: striketrough, ?

    @classmethod
    def from_json(cls, data: str | dict | list) -> Self:
        """
        Load from EditorJS JSON Blocks
        """
        data = data if isinstance(data, (dict, list)) else json.loads(data)
        blocks = data["blocks"] if isinstance(data, dict) else data
        markdown_items = []
        for child in blocks:
            _type = child["type"]
            if not (block := BLOCKS.get(_type)):
                warnings.warn(
                    f"from_json: Unsupported block type `{_type}`",
                    category=RuntimeWarning,
                )
                continue

            try:
                data = child.get("data", {})
                # forward any 'tunes' via data:
                data["tunes"] = data.get("tunes") or child.get("tunes") or {}

                markdown_items.append(block.to_markdown(data))
            except Exception as e:
                warnings.warn(
                    "from_json: Oh oh, unexpected block failure!",
                    category=RuntimeWarning,
                    source=e,
                )
                # if isinstance(e, TODO):
                #     raise e

        markdown = "".join(markdown_items)
        return cls.from_markdown(markdown)

    @classmethod
    def from_markdown(cls, data: str) -> Self:
        """
        Load from markdown string
        """

        return cls(mdast.md_to_json(data))

    @classmethod
    def from_mdast(cls, data: str | dict) -> Self:
        """
        Existing mdast representation
        """
        return cls(data)

    def to_json(self) -> str:
        """
        Export EditorJS JSON Blocks
        """
        # logic based on https://github.com/carrara88/editorjs-md-parser/blob/main/src/MarkdownImporter.js
        blocks = []
        for child in self._mdast["children"]:
            _type = child["type"]
            if not (block := BLOCKS.get(_type)):
                warnings.warn(
                    f"to_json: Unsupported block type `{_type}`",
                    category=RuntimeWarning,
                )
                continue

            try:
                blocks.extend(block.to_json(child))
            except Exception as e:

                warnings.warn(
                    "to_json: Oh oh, unexpected block failure!",
                    category=RuntimeWarning,
                    source=e,
                )
                raise e
                # if isinstance(e, TODO):
                #     raise e

        data = {"time": unix_timestamp(), "blocks": blocks, "version": EDITORJS_VERSION}

        return json.dumps(data)

    def to_markdown(self) -> str:
        """
        Export Markdown string
        """
        md = mdast.json_to_md(self.to_mdast())
        # idk why this happens:
        md = md.replace(r"\[ ]", "[ ]")
        md = md.replace(r"\[x]", "[x]")
        return md.replace(
            "  <", " <"
        )  # replace double space (added by mdast) with single one

    def to_mdast(self) -> str:
        """
        Export mdast representation
        """
        return json.dumps(self._mdast)

    def to_html(self) -> str:
        """
        Export HTML string
        """
        md = self.to_markdown()
        return self._md.convert(md)

    def __repr__(self):
        md = self.to_markdown()
        md = md.replace("\n", "\\n")
        return f"EditorJS({md})"

    def __str__(self):
        return self.to_markdown()

    # def __eq__(self, other: Self) -> bool:
    #     a = self.to_markdown()
    #     b = other.to_markdown()
    #
    #     remove = string.punctuation + string.whitespace
    #     return a.translate(remove) == b.translate(remove)
